# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Iterable, Optional, List, Set, Tuple, Union
from lale.search.search_space_grid import get_search_space_grids, SearchSpaceGrid
from lale.sklearn_compat import make_sklearn_compat, unnest_HPparams
import lale.operators as Ops
from lale.schema_utils import Schema, getMinimum, getMaximum
from lale.search.search_space import SearchSpace, SearchSpaceObject, SearchSpaceEnum, SearchSpaceNumber, SearchSpaceArray, should_print_search_space
from lale.search.PGO import PGO

import numpy as np
import sklearn.model_selection

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from lale.operators import PlannedOperator

def get_defaults_as_param_grid(op:'Ops.IndividualOp'):
    defaults = op.hyperparam_defaults()
    return {k : [v] for k,v in defaults.items()}

def get_lale_gridsearchcv_op(op, params, **kwargs):
    g = sklearn.model_selection.GridSearchCV(
        make_sklearn_compat(op), params, **kwargs)
    return g
#    TODO: turn it into a Lale TrainableOperator
#    name = f"GridSearchCV[{op.name()}]"
#    return Ops.TrainableIndividualOp(_name=name, _impl=g, _schemas=None)


def get_parameter_grids(
    op:'PlannedOperator', 
    num_samples:Optional[int]=None, 
    num_grids:Optional[float]=None,
    pgo:Optional[PGO]=None):
    """
    Parameters
    ----------
    op : The lale PlannedOperator
    lale_num_samples : integer, optional
        If set, will limit the number of samples for each distribution
    lale_num_grids: integer or float, optional
        if set to an integer => 1, it will determine how many parameter grids will be returned (at most)
        if set to an float between 0 and 1, it will determine what fraction should be returned
        note that setting it to 1 is treated as in integer.  To return all results, use None
    """
    return get_grid_search_parameter_grids(op, num_samples=num_samples, num_grids=num_grids, pgo=pgo)

def get_grid_search_parameter_grids(
        op:'PlannedOperator', 
        num_samples:Optional[int]=None, 
        num_grids:Optional[float]=None,
        pgo:Optional[PGO]=None)->List[Dict[str, List[Any]]]:
    """ Top level function: given a lale operator, returns a list of parameter grids
        suitable for passing to GridSearchCV.
        Note that you will need to wrap the lale operator for sklearn compatibility to call GridSearchCV
        directly.  The lale GridSearchCV wrapper takes care of that for you
    """
    hp_grids = get_search_space_grids(op, num_grids=num_grids, pgo=pgo)
    grids = SearchSpaceGridstoGSGrids(hp_grids, num_samples=num_samples)
    if should_print_search_space("true", "all", "backend", "gridsearchcv"):
        name = op.name()
        if not name:
            name = "an operator"
        print(f"GridSearchCV grids for {name}:\n{gridsearchcv_grids_to_string(grids)}")

    return grids

GSValue = Any
GSGrid = Dict[str,List[GSValue]]

DEFAULT_SAMPLES_PER_DISTRIBUTION=2

def SearchSpaceNumberToGSValues( key:str, 
                        hp:SearchSpaceNumber, 
                        num_samples:Optional[int]=None
                        )->List[GSValue]:
    """Returns either a list of values intended to be sampled uniformly """
    samples:int
    if num_samples is None:
        samples = DEFAULT_SAMPLES_PER_DISTRIBUTION
    else:
        samples = num_samples

    # Add preliminary support for PGO
    if hp.pgo is not None:
        ret = list(hp.pgo.samples(samples))
        return ret

    # if we are not doing PGO
    dist = "uniform"
    if hp.distribution:
        dist = hp.distribution    
    if hp.maximum is None:
        raise ValueError(f"maximum not specified for a number with distribution {dist} for {key}")
    max = hp.getInclusiveMax()
    if hp.minimum is None:
        raise ValueError(f"minimum not specified for a number with distribution {dist} for {key}")
    min = hp.getInclusiveMin()

    dt:np.dtype
    if hp.discrete:
        dt = np.dtype(int)
    else:
        dt = np.dtype(float)

    default = hp.default()
    if default is not None:
        # always use the default as one of the samples
        # TODO: ensure that the default is valid according to the schema
        if samples <= 1:
            return [default]
        samples = samples - 1
    if dist == "uniform" or dist == "integer":
        ret = np.linspace(min, max, num=samples, dtype=dt).tolist()
    elif dist == "loguniform":
        ret = np.logspace(min, max, num=samples, dtype=dt).tolist()
    else:
        raise ValueError(f"unknown/unsupported distribution {dist} for {key}")
    if default is not None:
        ret.append(default)
    return ret

def HPValuetoGSValue(key:str, 
                     hp:SearchSpace, 
                     num_samples:Optional[int]=None)->List[GSValue]:
    if isinstance(hp, SearchSpaceEnum):
        return hp.vals
    elif isinstance(hp, SearchSpaceNumber):
        return SearchSpaceNumberToGSValues(key, hp, num_samples=num_samples)
    elif isinstance(hp, SearchSpaceArray):
        raise ValueError(f"Arrays are not yet supported by the GridSearchCV backend (key: {key})")
    else:
        raise ValueError(f"Not yet supported hp description ({type(hp)}) (key: {key}) in the GridSearchCV backend")

def SearchSpaceGridtoGSGrid( hp:SearchSpaceGrid,
                    num_samples:Optional[int]=None)->GSGrid:
    return {k:HPValuetoGSValue(k, v, num_samples=num_samples) for k,v in hp.items()}

def SearchSpaceGridstoGSGrids(hp_grids:List[SearchSpaceGrid],
                     num_samples:Optional[int]=None)->List[GSGrid]:
    return [SearchSpaceGridtoGSGrid(g, num_samples=num_samples) for g in hp_grids]

def gridsearchcv_grid_to_string(grid:GSGrid)->str:
    return "{" + ";".join(f"{k}->{str(v)}" for k,v in grid.items()) + "}"

def gridsearchcv_grids_to_string(grids:List[GSGrid])->str:
    return "|".join(gridsearchcv_grid_to_string(grid) for grid in grids)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.docstrings
import lale.operators
import sklearn.compose

class ColumnTransformerImpl:
    def __init__(self, **hyperparams):
        self._wrapped_model = sklearn.compose.ColumnTransformer(**hyperparams)

    def fit(self, X, y=None):
        self._wrapped_model.fit(X, y)
        return self

    def transform(self, X):
        result = self._wrapped_model.transform(X)
        return result

_hyperparams_schema = {
  'allOf': [
  { 'description':
      'This first sub-object lists all constructor arguments with their '
      'types, one at a time, omitting cross-argument constraints, if any.',
    'type': 'object',
    'additionalProperties': False,
    'required': ['transformers'],
    'relevantToOptimizer': [],
    'properties': {
      'transformers': {
        'description':
          'Operators or pipelines to be applied to subsets of the data.',
        'type': 'array',
        'items': {
          'description': 'Tuple of (name, transformer, column(s)).',
          'type': 'array',
          'laleType': 'tuple',
          'minItems': 3, 'maxItems': 3,
          'items': [
          { 'description': 'Name.',
            'type': 'string' },
          { 'description': 'Transformer.',
            'anyOf': [
            { 'description': 'Transformer supporting fit and transform.',
              'laleType': 'operator'},
            { 'enum': ['passthrough', 'drop']}]},
          { 'description': 'Column(s).',
            'anyOf': [
             { 'type': 'integer',
               'description': 'One column by index.'},
             { 'type': 'array', 'items': {'type': 'integer'},
               'description': 'Multiple columns by index.'},
             { 'type': 'string',
               'description': 'One Dataframe column by name.'},
             { 'type': 'array', 'items': {'type': 'string'},
               'description': 'Multiple Dataframe columns by names.'},
             { 'type': 'array', 'items': {'type': 'boolean'},
               'description': 'Boolean mask.'}]}]}},
      'remainder': {
        'description':
          'Transformation for columns that were not specified in transformers.',
        'anyOf': [
        { 'description': 'Transformer supporting fit and transform.',
          'laleType': 'operator'},
        { 'enum': ['passthrough', 'drop']}],
        'default': 'drop'},
      'sparse_threshold': {
        'description': """If the output of the different transfromers contains sparse matrices,
these will be stacked as a sparse matrix if the overall density is
lower than this value. Use sparse_threshold=0 to always return dense.""",
        'type': 'number',
        'default': 0.3},  
      'n_jobs': {
        'description': 'Number of jobs to run in parallel',
        'anyOf': [
          { 'description': '1 unless in joblib.parallel_backend context.',
            'enum': [None]},
          { 'description': 'Use all processors.',
            'enum': [-1]},
          { 'description': 'Number of CPU cores.',
            'type': 'integer',
            'minimum': 1}],
          'default': None},
      'transformer_weights': {
        'description': """Multiplicative weights for features per transformer.
The output of the transformer is multiplied by these weights.""",
        'anyOf': [
        { 'description': 'Keys are transformer names, values the weights.',
          'type': 'object'},
        { 'enum': [None]}],
        'default': None}}}]}

_input_fit_schema = {
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {
        'type': 'array',
        'items': {
          'anyOf':[{'type': 'number'}, {'type':'string'}]}}},
    'y': {
      'description': 'Target for supervised learning (ignored).'}}}

_input_transform_schema = {
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {
        'type': 'array',
        'items': {
           'anyOf':[{'type': 'number'}, {'type':'string'}]}}}}}

_output_transform_schema = {
  'description': 'Features; the outer array is over samples.',
  'type': 'array',
  'items': {
    'type': 'array',
    'items': {'type': 'number'}}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """ColumnTransformer_ applies transformers to columns of an array or pandas DataFrame.

.. _ColumnTransformer: https://scikit-learn.org/0.20/modules/generated/sklearn.compose.ColumnTransformer.html#sklearn-compose-columntransformer
""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.column_transformer.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(ColumnTransformerImpl, _combined_schemas)

ColumnTransformer = lale.operators.make_operator(ColumnTransformerImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Scikit-learn compatible wrappers for a subset of the operators from AIF360_ along with schemas to enable hyperparameter tuning.

.. _AIF360: https://github.com/IBM/AIF360

Operators:
==========
* `CalibratedEqOddsPostprocessing`_
* `DisparateImpactRemover`_
* `EqOddsPostprocessing`_
* `RejectOptionClassification`_

.. _`CalibratedEqOddsPostprocessing`: lale.lib.aif360.calibrated_eq_odds_postprocessing.html
.. _`DisparateImpactRemover`: lale.lib.aif360.disparate_impact_remover.html
.. _`EqOddsPostprocessing`: lale.lib.aif360.eq_odds_postprocessing.html
.. _`RejectOptionClassification`: lale.lib.aif360.reject_option_classification.html

Functions:
==========
* `dataset_fairness_info`_
* `dataset_to_pandas`_
* `disparate_impact`_
* `statistical_parity_difference`_

.. _`dataset_fairness_info`: lale.lib.aif360.util.html#lale.lib.aif360.util.dataset_fairness_info
.. _`dataset_to_pandas`: lale.lib.aif360.util.html#lale.lib.aif360.util.dataset_to_pandas
.. _`disparate_impact`: lale.lib.aif360.util.html#lale.lib.aif360.util.disparate_impact
.. _`statistical_parity_difference`: lale.lib.aif360.util.html#lale.lib.aif360.util.statistical_parity_difference
"""

from .calibrated_eq_odds_postprocessing import CalibratedEqOddsPostprocessing
from .disparate_impact_remover import DisparateImpactRemover
from .eq_odds_postprocessing import EqOddsPostprocessing
from .reject_option_classification import RejectOptionClassification
from .util import dataset_fairness_info
from .util import dataset_to_pandas
from .util import disparate_impact
from .util import statistical_parity_difference




from sklearn.kernel_approximation import RBFSampler as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class RBFSamplerImpl():

    def __init__(self, gamma=1.0, n_components=100, random_state=None):
        self._hyperparams = {
            'gamma': gamma,
            'n_components': n_components,
            'random_state': random_state}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for RBFSampler    Approximates feature map of an RBF kernel by Monte Carlo approximation',
    'allOf': [{
        'type': 'object',
        'required': ['gamma', 'n_components', 'random_state'],
        'relevantToOptimizer': ['n_components'],
        'additionalProperties': False,
        'properties': {
            'gamma': {
                'type': 'number',
                'default': 1.0,
                'description': 'Parameter of RBF kernel: exp(-gamma * x^2)'},
            'n_components': {
                'type': 'integer',
                'minimumForOptimizer': 2,
                'maximumForOptimizer': 256,
                'distribution': 'uniform',
                'default': 100,
                'description': 'Number of Monte Carlo samples per original feature'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`.'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the model with X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training data, where n_samples in the number of samples and n_features is the number of features.'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Apply the approximate feature map to X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'New data, where n_samples in the number of samples and n_features is the number of features.'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Apply the approximate feature map to X.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.kernel_approximation.RBFSampler#sklearn-kernel_approximation-rbfsampler',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(RBFSamplerImpl, _combined_schemas)
RBFSampler = lale.operators.make_operator(RBFSamplerImpl, _combined_schemas)





from sklearn.preprocessing.data import Binarizer as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class BinarizerImpl():

    def __init__(self, threshold=0.0, copy=True):
        self._hyperparams = {
            'threshold': threshold,
            'copy': copy}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for Binarizer    Binarize data (set feature values to 0 or 1) according to a threshold',
    'allOf': [{
        'type': 'object',
        'required': ['threshold', 'copy'],
        'relevantToOptimizer': ['copy'],
        'additionalProperties': False,
        'properties': {
            'threshold': {
                'XXX TODO XXX': 'float, optional (0.0 by default)',
                'description': 'Feature values below or equal to this are replaced by 0, above it by 1',
                'type': 'number',
                'default': 0.0},
            'copy': {
                'type': 'boolean',
                'default': True,
                'description': 'set to False to perform inplace binarization and avoid a copy (if the input is already a numpy array or a scipy.sparse CSR matrix).'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Do nothing and return the estimator unchanged',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'laleType': 'Any',
                'XXX TODO XXX': 'item type'},
            'XXX TODO XXX': 'array-like'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Binarize each element of X',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The data to binarize, element by element'},
        'y': {
            'laleType': 'Any',
            'XXX TODO XXX': '(ignored)',
            'description': ''},
        'copy': {
            'type': 'boolean',
            'description': 'Copy the input X or not.'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Binarize each element of X',
    'laleType': 'Any',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.Binarizer#sklearn-preprocessing-binarizer',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(BinarizerImpl, _combined_schemas)
Binarizer = lale.operators.make_operator(BinarizerImpl, _combined_schemas)





from sklearn.decomposition.sparse_pca import MiniBatchSparsePCA as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class MiniBatchSparsePCAImpl():

    def __init__(self, n_components=None, alpha=1, ridge_alpha=0.01, n_iter=100, callback=None, batch_size=3, verbose=False, shuffle=True, n_jobs=None, method='lars', random_state=None, normalize_components=False):
        self._hyperparams = {
            'n_components': n_components,
            'alpha': alpha,
            'ridge_alpha': ridge_alpha,
            'n_iter': n_iter,
            'callback': callback,
            'batch_size': batch_size,
            'verbose': verbose,
            'shuffle': shuffle,
            'n_jobs': n_jobs,
            'method': method,
            'random_state': random_state,
            'normalize_components': normalize_components}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for MiniBatchSparsePCA    Mini-batch Sparse Principal Components Analysis',
    'allOf': [{
        'type': 'object',
        'required': ['n_components', 'alpha', 'ridge_alpha', 'n_iter', 'callback', 'batch_size', 'verbose', 'shuffle', 'n_jobs', 'method', 'random_state', 'normalize_components'],
        'relevantToOptimizer': ['n_components', 'alpha', 'n_iter', 'batch_size', 'shuffle', 'method'],
        'additionalProperties': False,
        'properties': {
            'n_components': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 2,
                    'maximumForOptimizer': 256,
                    'distribution': 'uniform'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'number of sparse atoms to extract'},
            'alpha': {
                'type': 'integer',
                'minimumForOptimizer': 1,
                'maximumForOptimizer': 2,
                'distribution': 'uniform',
                'default': 1,
                'description': 'Sparsity controlling parameter'},
            'ridge_alpha': {
                'type': 'number',
                'default': 0.01,
                'description': 'Amount of ridge shrinkage to apply in order to improve conditioning when calling the transform method.'},
            'n_iter': {
                'type': 'integer',
                'minimumForOptimizer': 5,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 100,
                'description': 'number of iterations to perform for each mini batch'},
            'callback': {
                'anyOf': [{
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'callable that gets invoked every five iterations'},
            'batch_size': {
                'type': 'integer',
                'minimumForOptimizer': 3,
                'maximumForOptimizer': 128,
                'distribution': 'uniform',
                'default': 3,
                'description': 'the number of features to take in each mini batch'},
            'verbose': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'boolean'}],
                'default': False,
                'description': 'Controls the verbosity; the higher, the more messages'},
            'shuffle': {
                'type': 'boolean',
                'default': True,
                'description': 'whether to shuffle the data before splitting it in batches'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Number of parallel jobs to run'},
            'method': {
                'enum': ['lars', 'cd'],
                'default': 'lars',
                'description': 'lars: uses the least angle regression method to solve the lasso problem (linear_model.lars_path) cd: uses the coordinate descent method to compute the Lasso solution (linear_model.Lasso)'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`.'},
            'normalize_components': {
                'type': 'boolean',
                'default': False,
                'description': '- if False, use a version of Sparse PCA without components   normalization and without data centering'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the model from data in X.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vector, where n_samples in the number of samples and n_features is the number of features.'},
        'y': {
            
        }},
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Least Squares projection of the data onto the sparse components.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Test data to be transformed, must have the same number of features as the data used to train the model.'},
        'ridge_alpha': {
            'type': 'number',
            'default': 0.01,
            'description': 'Amount of ridge shrinkage to apply in order to improve conditioning'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transformed data.',
    'laleType': 'Any',
    'XXX TODO XXX': 'X_new array, shape (n_samples, n_components)',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.decomposition.MiniBatchSparsePCA#sklearn-decomposition-minibatchsparsepca',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(MiniBatchSparsePCAImpl, _combined_schemas)
MiniBatchSparsePCA = lale.operators.make_operator(MiniBatchSparsePCAImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Schema-enhanced versions of a subset of the operators from `autoai_libs`_ to enable hyperparameter tuning.

.. _`autoai_libs`: https://pypi.org/project/autoai-libs

Operators
=========


* lale.lib.autoai_libs. `NumpyColumnSelector`_
* lale.lib.autoai_libs. `CompressStrings`_
* lale.lib.autoai_libs. `NumpyReplaceMissingValues`_
* lale.lib.autoai_libs. `NumpyReplaceUnknownValues`_
* lale.lib.autoai_libs. `boolean2float`_
* lale.lib.autoai_libs. `CatImputer`_
* lale.lib.autoai_libs. `float32_transform`_
* lale.lib.autoai_libs. `FloatStr2Float`_
* lale.lib.autoai_libs. `NumImputer`_
* lale.lib.autoai_libs. `OptStandardScaler`_
* lale.lib.autoai_libs. `NumpyPermuteArray`_
* lale.lib.autoai_libs. `TA1`_
* lale.lib.autoai_libs.`CatEncoder`_

.. _`NumpyColumnSelector`: lale.lib.autoai_libs.numpy_column_selector.html
.. _`CompressStrings`: lale.lib.autoai_libs.compress_strings.html
.. _`NumpyReplaceMissingValues`: lale.lib.autoai_libs.numpy_replace_missing_values.html
.. _`NumpyReplaceUnknownValues`: lale.lib.autoai_libs.numpy_replace_unknown_values.html
.. _`boolean2float`: lale.lib.autoai_libs.boolean2float.html
.. _`CatImputer`: lale.lib.autoai_libs.cat_imputer.html
.. _`float32_transform`: lale.lib.autoai_libs.float32_transform.html
.. _`FloatStr2Float`: lale.lib.autoai_libs.float_str2_float.html
.. _`NumImputer`: lale.lib.autoai_libs.num_imputer.html
.. _`OptStandardScaler`: lale.lib.autoai_libs.opt_standard_scaler.html
.. _`NumpyPermuteArray`: lale.lib.autoai_libs.numpy_permute_array.html
.. _`TA1`: lale.lib.autoai_libs.ta1.html
.. _`CatEncoder`: lale.lib.autoai_libs.cat_encoder.html
"""

# from autoai_libs.transformers.exportable
from .numpy_column_selector        import NumpyColumnSelector
from .compress_strings             import CompressStrings
from .numpy_replace_missing_values import NumpyReplaceMissingValues
from .numpy_replace_unknown_values import NumpyReplaceUnknownValues
from .boolean2float                import boolean2float
from .cat_imputer                  import CatImputer
from .cat_encoder                  import CatEncoder
from .float32_transform            import float32_transform
from .float_str2_float             import FloatStr2Float
from .num_imputer                  import NumImputer
from .opt_standard_scaler          import OptStandardScaler
from .numpy_permute_array          import NumpyPermuteArray

# from autoai_libs.cognito.transforms.transform_utils
from .ta1  import TA1
# from .ta2  import TA2
# from .tb1  import TB1
# from .tb2  import TB2
# from .tam  import TAM
# from .tgen import TGen
# from .fs1  import FS1
# from .fs2  import FS2



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.preprocessing
import lale.operators

class FunctionTransformerImpl():
    def __init__(self, func=None, inverse_func=None, validate=None, accept_sparse=False, pass_y='deprecated', check_inverse=True, kw_args=None, inv_kw_args=None):
        self._hyperparams = {
            'func': func,
            'inverse_func': inverse_func,
            'validate': validate,
            'accept_sparse': accept_sparse,
            'pass_y': pass_y,
            'check_inverse': check_inverse,
            'kw_args': kw_args,
            'inv_kw_args': inv_kw_args}
        self._wrapped_model = sklearn.preprocessing.FunctionTransformer(
            **self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)

_hyperparams_schema = {
    'allOf': [
    {   'description': 'This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.\n',
        'type': 'object',
        'additionalProperties': False,
        'required': ['func', 'inverse_func', 'validate', 'accept_sparse', 'pass_y', 'check_inverse', 'kw_args', 'inv_kw_args'],
        'relevantToOptimizer': [],
        'properties': {
            'func': {
                'anyOf': [
                {   'laleType': 'Any'},
                {   'enum': [None]}],
                'default': None,
                'description': 'The callable to use for the transformation.'},
            'inverse_func': {
                'anyOf': [
                {   'laleType': 'Any'},
                {   'enum': [None]}],
                'default': None,
                'description': 'The callable to use for the inverse transformation.'},
            'validate': {
                'default': False,
                'description': 'Indicate that the input X array should be checked before calling ``func``.'},
            'accept_sparse': {
                'type': 'boolean',
                'default': False,
                'description': 'Indicate that func accepts a sparse matrix as input.'},
            'pass_y': {
                'anyOf': [{
                    'type': 'boolean'}, {
                    'enum': ['deprecated']}],
                'default': 'deprecated',
                'description': 'Indicate that transform should forward the y argument to the inner callable.'},
            'check_inverse': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether to check that or ``func`` followed by ``inverse_func`` leads to the original inputs.'},
            'kw_args': {
                'anyOf': [
                {   'type': 'object'},
                {   'enum': [None]}],
                'default': None,
                'description': 'Dictionary of additional keyword arguments to pass to func.'},
            'inv_kw_args': {
                'anyOf': [
                {   'type': 'object'},
                {   'enum': [None]}],
                'default': None,
                'description': 'Dictionary of additional keyword arguments to pass to inverse_func.'}}},
    {   'description': 'If validate is False, then accept_sparse has no effect.',
        'anyOf': [
        {   'type': 'object',
            'properties': {'validate': {'not': {'enum': [False]}}}},
        {   'type': 'object',
            'properties': {'accept_sparse': {'enum': [False]}}}]}]}

_input_fit_schema = {
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'anyOf':[{'type': 'number'}, {'type':'string'}]}}},
        'y': {'laleType': 'Any'}}}

_input_transform_schema = {
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'anyOf':[{'type': 'number'}, {'type':'string'}]}}}}}

_output_transform_schema = {
    'type': 'array',
    'items': {'laleType': 'Any'}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """FunctionTransformer_ constructs a transformer from an arbitrary callable.

.. _FunctionTransformer: https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.FunctionTransformer.html
""",
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.FunctionTransformer.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(FunctionTransformerImpl, _combined_schemas)

FunctionTransformer = lale.operators.make_operator(FunctionTransformerImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.datasets.data_schemas
import lale.docstrings
import lale.operators
import lale.type_checking
import sklearn.compose
import pandas as pd

class ProjectImpl:
    def __init__(self, columns=None):
        self._hyperparams = { 'columns': columns }

    def fit(self, X, y=None):
        columns = self._hyperparams['columns']
        if lale.type_checking.is_schema(columns):
            s_all = lale.datasets.data_schemas.to_schema(X)
            s_row = s_all['items']
            n_columns = s_row['minItems']
            assert n_columns == s_row['maxItems']
            s_cols = s_row['items']
            if isinstance(s_cols, dict):
                if lale.type_checking.is_subschema(s_cols, columns):
                    columns = [*range(n_columns)]
                else:
                    columns = []
            else:
                assert isinstance(s_cols, list)
                columns = [
                    i for i in range(n_columns)
                    if lale.type_checking.is_subschema(s_cols[i], columns)]
        self._col_tfm = sklearn.compose.ColumnTransformer(
            transformers=[('keep', 'passthrough', columns)])
        self._col_tfm.fit(X)
        return self

    def transform(self, X):
        columns = self._col_tfm.transformers[0][2]
        if isinstance(X, pd.DataFrame) and isinstance(columns, list):
            col_names = [
                X.columns[c] if isinstance(c, int) else c for c in columns]
            result = X[col_names]
        else:
            result = self._col_tfm.transform(X)
        s_X = lale.datasets.data_schemas.to_schema(X)
        s_result = self.transform_schema(s_X)
        return lale.datasets.data_schemas.add_schema(result, s_result)

    def transform_schema(self, s_X):
        """Used internally by Lale for type-checking downstream operators."""
        if hasattr(self, '_col_tfm'):
            return self._transform_schema_col_tfm(s_X, self._col_tfm)
        columns = self._hyperparams['columns']
        if lale.type_checking.is_schema(columns):
            return self._transform_schema_schema(s_X, columns)
        if not lale.type_checking.is_schema(s_X):
            X = lale.datasets.data_schemas.add_schema(s_X)
            self.fit(X)
            return self._transform_schema_col_tfm(X.json_schema, self._col_tfm)
        return s_X

    def _transform_schema_col_tfm(self, s_X, col_tfm):
        s_X = lale.datasets.data_schemas.to_schema(s_X)
        s_row = s_X['items']
        s_cols = s_row['items']
        keep_cols = [col
                     for name, tfm, cols in col_tfm.transformers_
                     if tfm == 'passthrough'
                     for col in cols]
        n_columns = len(keep_cols)
        if isinstance(s_cols, dict):
            s_cols_result = s_cols
        else:
            name2i = {s_cols[i]['description']: i for i in range(len(s_cols))}
            keep_cols_i = [name2i[col] if isinstance(col, str) else col
                           for col in keep_cols]
            s_cols_result = [s_cols[i] for i in keep_cols_i]
        s_result = {
            **s_X,
            'items': {
                **s_row,
                'minItems': n_columns, 'maxItems': n_columns,
                'items': s_cols_result}}
        return s_result

    def _transform_schema_schema(self, s_X, schema):
        s_X = lale.datasets.data_schemas.to_schema(s_X)
        s_row = s_X['items']
        s_cols = s_row['items']
        if isinstance(s_cols, dict):
            if lale.type_checking.is_subschema(s_cols, schema):
                s_row_result = s_row
            else:
                s_row_result = {'type': 'array', 'minItems': 0, 'maxItems': 0}
        else:
            assert isinstance(s_cols, list)
            s_cols_result = [s for s in s_cols
                             if lale.type_checking.is_subschema(s, schema)]
            n_columns = len(s_cols_result)
            s_row_result = {
                'type': 'array',
                'minItems': n_columns, 'maxItems': n_columns,
                'items': s_cols_result}
        return {'type': 'array', 'items': s_row_result}

_hyperparams_schema = {
  'allOf': [
    { 'description':
        'This first sub-object lists all constructor arguments with their '
        'types, one at a time, omitting cross-argument constraints, if any.',
      'type': 'object',
      'additionalProperties': False,
      'required': ['columns'],
      'relevantToOptimizer': [],
      'properties': {
          'columns': {
              'description': """The subset of columns to retain.

The supported column specification formats include those of
scikit-learn's ColumnTransformer_, and in addition, filtering
by using a JSON subschema_ check.

.. _ColumnTransformer: https://scikit-learn.org/stable/modules/generated/sklearn.compose.ColumnTransformer.html
.. _subschema: https://github.com/IBM/jsonsubschema""",
              'anyOf': [
                 { 'type': 'integer',
                   'description': 'One column by index.'},
                 { 'type': 'array', 'items': {'type': 'integer'},
                   'description': 'Multiple columns by index.'},
                 { 'type': 'string',
                   'description': 'One Dataframe column by name.'},
                 { 'type': 'array', 'items': {'type': 'string'},
                   'description': 'Multiple Dataframe columns by names.'},
                 { 'type': 'array', 'items': {'type': 'boolean'},
                   'description': 'Boolean mask.'},
                 { 'type': 'object',
                   'description': 'Keep columns whose schema is a subschema of this JSON schema.'}]}}}]}

_input_fit_schema = {
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {
        'type': 'array',
        'items': {
          'anyOf':[{'type': 'number'}, {'type':'string'}]}}},
    'y': {
      'description': 'Target for supervised learning (ignored).'}}}

_input_transform_schema = {
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {
        'type': 'array',
        'items': {
           'anyOf':[{'type': 'number'}, {'type':'string'}]}}}}}

_output_transform_schema = {
  'description': 'Features; the outer array is over samples.',
  'type': 'array',
  'items': {
    'type': 'array',
    'items': {
       'anyOf':[{'type': 'number'}, {'type':'string'}]}}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """Projection keeps a subset of the columns, like in relational algebra.

Examples
--------
>>> df = pd.DataFrame(data={'A': [1,2], 'B': ['x','y'], 'C': [3,4]})
>>> keep_numbers = Project(columns={'type': 'number'})
>>> keep_numbers.fit(df).transform(df)
NDArrayWithSchema([[1, 3],
                   [2, 4]])
""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.project.html',
    'type': 'object',
    'tags': {
        'pre': ['categoricals'],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(ProjectImpl, _combined_schemas)

Project = lale.operators.make_operator(ProjectImpl, _combined_schemas)




from sklearn.calibration import CalibratedClassifierCV as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class CalibratedClassifierCVImpl():

    def __init__(self, base_estimator=None, method='sigmoid', cv=3):
        self._hyperparams = {
            'base_estimator': base_estimator,
            'method': method,
            'cv': cv}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for CalibratedClassifierCV    Probability calibration with isotonic regression or sigmoid.',
    'allOf': [{
        'type': 'object',
        'required': ['base_estimator', 'method', 'cv'],
        'relevantToOptimizer': ['method', 'cv'],
        'additionalProperties': False,
        'properties': {
            'base_estimator': {
                'XXX TODO XXX': 'instance BaseEstimator',
                'description': 'The classifier whose output decision function needs to be calibrated to offer more accurate predict_proba outputs',
                'enum': [None],
                'default': None},
            'method': {
                'XXX TODO XXX': "'sigmoid' or 'isotonic'",
                'description': 'The method to use for calibration',
                'enum': ['isotonic', 'sigmoid'],
                'default': 'sigmoid'},
            'cv': {
                'XXX TODO XXX': 'integer, cross-validation generator, iterable or "prefit", optional',
                'description': 'Determines the cross-validation splitting strategy',
                'type': 'integer',
                'minimumForOptimizer': 3,
                'maximumForOptimizer': 4,
                'distribution': 'uniform',
                'default': 3},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the calibrated model',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training data.'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target values.'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'description': 'Sample weights'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict the target of new samples. Can be different from the',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The samples.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The predicted class.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_input_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Posterior probabilities of classification',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The samples.'},
    },
}
_output_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The predicted probas.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.calibration.CalibratedClassifierCV#sklearn-calibration-calibratedclassifiercv',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_predict_proba': _input_predict_proba_schema,
        'output_predict_proba': _output_predict_proba_schema},
}
lale.docstrings.set_docstrings(CalibratedClassifierCVImpl, _combined_schemas)
CalibratedClassifierCV = lale.operators.make_operator(CalibratedClassifierCVImpl, _combined_schemas)





from sklearn.manifold.isomap import Isomap as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class IsomapImpl():

    def __init__(self, n_neighbors=5, n_components=2, eigen_solver='auto', tol=0, max_iter=None, path_method='auto', neighbors_algorithm='auto', n_jobs=None):
        self._hyperparams = {
            'n_neighbors': n_neighbors,
            'n_components': n_components,
            'eigen_solver': eigen_solver,
            'tol': tol,
            'max_iter': max_iter,
            'path_method': path_method,
            'neighbors_algorithm': neighbors_algorithm,
            'n_jobs': n_jobs}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for Isomap    Isomap Embedding',
    'allOf': [{
        'type': 'object',
        'required': ['n_neighbors', 'n_components', 'eigen_solver', 'tol', 'max_iter', 'path_method', 'neighbors_algorithm', 'n_jobs'],
        'relevantToOptimizer': ['n_neighbors', 'n_components', 'eigen_solver', 'tol', 'path_method', 'neighbors_algorithm'],
        'additionalProperties': False,
        'properties': {
            'n_neighbors': {
                'type': 'integer',
                'minimumForOptimizer': 5,
                'maximumForOptimizer': 20,
                'distribution': 'uniform',
                'default': 5,
                'description': 'number of neighbors to consider for each point.'},
            'n_components': {
                'type': 'integer',
                'minimumForOptimizer': 2,
                'maximumForOptimizer': 256,
                'distribution': 'uniform',
                'default': 2,
                'description': 'number of coordinates for the manifold'},
            'eigen_solver': {
                'enum': ['auto', 'arpack', 'dense'],
                'default': 'auto',
                'description': "'auto' : Attempt to choose the most efficient solver for the given problem"},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 0.0,
                'maximumForOptimizer': 1.0,
                'distribution': 'uniform',
                'default': 0,
                'description': 'Convergence tolerance passed to arpack or lobpcg'},
            'max_iter': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Maximum number of iterations for the arpack solver'},
            'path_method': {
                'enum': ['auto', 'FW', 'D'],
                'default': 'auto',
                'description': 'Method to use in finding shortest path'},
            'neighbors_algorithm': {
                'enum': ['auto', 'brute', 'kd_tree', 'ball_tree'],
                'default': 'auto',
                'description': 'Algorithm to use for nearest neighbors search, passed to neighbors.NearestNeighbors instance.'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of parallel jobs to run'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Compute the embedding vectors for data X',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'laleType': 'Any',
                'XXX TODO XXX': 'item type'},
            'XXX TODO XXX': '{array-like, sparse matrix, BallTree, KDTree, NearestNeighbors}',
            'description': 'Sample data, shape = (n_samples, n_features), in the form of a numpy array, precomputed tree, or NearestNeighbors object.'},
        'y': {
            
        }},
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform X.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.manifold.Isomap#sklearn-manifold-isomap',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(IsomapImpl, _combined_schemas)
Isomap = lale.operators.make_operator(IsomapImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.operators
import lale.docstrings
import sklearn.preprocessing

class MinMaxScalerImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = sklearn.preprocessing.MinMaxScaler(
            **self._hyperparams)

    def fit(self, X, y=None):
        self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)

    def partial_fit(self, X, y=None):
      if not hasattr(self, "_wrapped_model"):
        self._wrapped_model = sklearn.preprocessing.MinMaxScaler(
            **self._hyperparams)
      self._wrapped_model.partial_fit(X)
      return self

_input_schema_fit = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Input data schema for training.',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}},
    'y': {}}}

_input_transform_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Input data schema for predictions.',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}}}}

_output_transform_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Output data schema for transformed data.',
  'type': 'array',
  'items': {'type': 'array', 'items': {'type': 'number'}}}

_hyperparams_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Hyperparameter schema.',
  'allOf': [
    { 'description':
        'This first sub-object lists all constructor arguments with their '
        'types, one at a time, omitting cross-argument constraints.',
      'type': 'object',
      'additionalProperties': False,
      'required': ['feature_range', 'copy'],
      'relevantToOptimizer': [], #['feature_range'],
      'properties': {
        'feature_range': {
          'description': 'Desired range of transformed data.',
          'type': 'array',
          'laleType': 'tuple',
          'minItems': 2,
          'maxItems': 2,
          'default': [0, 1]},
        'copy': {
          'description':
            'Set to False to perform inplace row normalization and avoid '
            'a copy (if the input is already a numpy array).',
          'type': 'boolean',
          'default': True}}}]}

_combined_schemas = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Combined schema for expected data and hyperparameters.',
  'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.MinMaxScaler.html#sklearn-preprocessing-minmaxscaler',
  'type': 'object',
  'tags': {
    'pre': ['~categoricals'],
    'op': ['transformer', 'interpretable'],
    'post': []},
  'properties': {
    'hyperparams': _hyperparams_schema,
    'input_fit': _input_schema_fit,
    'input_transform': _input_transform_schema,
    'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(MinMaxScalerImpl, _combined_schemas)

MinMaxScaler = lale.operators.make_operator(MinMaxScalerImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aif360.algorithms.postprocessing
import lale.docstrings
import lale.operators
from .util import _BasePostprocessingImpl
from .util import _numeric_supervised_input_fit_schema
from .util import _numeric_input_predict_schema
from .util import _numeric_output_predict_schema
from .util import _postprocessing_base_hyperparams
from .util import _dataset_fairness_properties

_additional_hyperparams = {
    'unprivileged_groups': _dataset_fairness_properties['unprivileged_groups'],
    'privileged_groups': _dataset_fairness_properties['privileged_groups'],
    'seed': {
        'description': 'Seed to make `predict` repeatable.',
        'anyOf': [
        {   'type': 'integer'},
        {   'enum': [None]}],
        'default': None}}

class EqOddsPostprocessingImpl(_BasePostprocessingImpl):
    def __init__(self, estimator, favorable_label, unfavorable_label, protected_attribute_names, unprivileged_groups, privileged_groups, seed):
        mitigator = aif360.algorithms.postprocessing.EqOddsPostprocessing(
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups,
            seed=seed)
        super(EqOddsPostprocessingImpl, self).__init__(
            mitigator=mitigator,
            estimator=estimator,
            favorable_label=favorable_label,
            unfavorable_label=unfavorable_label,
            protected_attribute_names=protected_attribute_names)

_input_fit_schema = _numeric_supervised_input_fit_schema
_input_predict_schema = _numeric_input_predict_schema
_output_predict_schema = _numeric_output_predict_schema

_hyperparams_schema = {
    'description': 'Hyperparameter schema.',
    'allOf': [
    {   'description':
            'This first sub-object lists all constructor arguments with their '
            'types, one at a time, omitting cross-argument constraints.',
        'type': 'object',
        'additionalProperties': False,
        'required': (list(_postprocessing_base_hyperparams.keys())
                     + list(_additional_hyperparams.keys())),
        'relevantToOptimizer': [],
        'properties': {**_postprocessing_base_hyperparams,
                       **_additional_hyperparams}}]}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """`Equalized odds postprocessing`_ for fairness mitigation.

.. _`Equalized odds postprocessing`: https://aif360.readthedocs.io/en/latest/modules/generated/aif360.algorithms.postprocessing.EqOddsPostprocessing.html
""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.aif360.eq_odds_postprocessing.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'classifier', 'interpretable'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
}}

lale.docstrings.set_docstrings(EqOddsPostprocessingImpl, _combined_schemas)

EqOddsPostprocessing = lale.operators.make_operator(EqOddsPostprocessingImpl, _combined_schemas)




from sklearn.neighbors.nearest_centroid import NearestCentroid as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class NearestCentroidImpl():

    def __init__(self, metric='euclidean', shrink_threshold=None):
        self._hyperparams = {
            'metric': metric,
            'shrink_threshold': shrink_threshold}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for NearestCentroid    Nearest centroid classifier.',
    'allOf': [{
        'type': 'object',
        'required': ['metric', 'shrink_threshold'],
        'relevantToOptimizer': ['metric'],
        'additionalProperties': False,
        'properties': {
            'metric': {
                'anyOf': [{
                    'type': 'object',
                    'forOptimizer': False}, {
                    'enum': ['euclidean', 'manhattan', 'minkowski']}],
                'default': 'euclidean',
                'description': 'The metric to use when calculating distance between instances in a feature array'},
            'shrink_threshold': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Threshold for shrinking centroids to remove features.'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the NearestCentroid model according to the given training data.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vector, where n_samples is the number of samples and n_features is the number of features'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target values (integers)'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Perform classification on an array of test vectors X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Perform classification on an array of test vectors X.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.neighbors.NearestCentroid#sklearn-neighbors-nearestcentroid',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema},
}
lale.docstrings.set_docstrings(NearestCentroidImpl, _combined_schemas)
NearestCentroid = lale.operators.make_operator(NearestCentroidImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import random
import math
import numpy
import warnings

from typing import Any, Dict, List, Set, Iterable, Iterator, Optional, Tuple
from hyperopt import hp
from hyperopt.pyll import scope
from lale.schema_simplifier import findRelevantFields, narrowToGivenRelevantFields, simplify, filterForOptimizer

from lale.search.search_space_grid import get_search_space_grids, SearchSpaceGrid
from lale.sklearn_compat import make_sklearn_compat

from lale.schema_utils import Schema, getMinimum, getMaximum
from lale.search.search_space import SearchSpace, SearchSpaceObject, SearchSpaceEnum, SearchSpaceNumber, SearchSpaceArray, should_print_search_space
from lale.search.PGO import PGO
import numpy as np

from smac.configspace import ConfigurationSpace
from ConfigSpace.hyperparameters import CategoricalHyperparameter, \
    UniformFloatHyperparameter, UniformIntegerHyperparameter, Hyperparameter
from ConfigSpace.conditions import EqualsCondition

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import lale.operators as Ops

def lale_op_smac_tae(op:'Ops.PlannedOperator', f_min):
    # TODO: we can probably do this in a different way, but get_smac_configuration_space
    # we already have these sklearn compatibility wrappers it is easier for now to use them
    op_compat = make_sklearn_compat(op)
    def f(cfg):
        from sklearn.base import clone
        wrapped_op = clone(op_compat)
        cfg2 = smac_fixup_params(cfg)
        trainable = wrapped_op.set_params(**cfg2)

        return f_min(trainable)
        
    return f

def lale_trainable_op_from_config(op:'Ops.PlannedOperator', cfg)->'Ops.TrainableOperator':
    from sklearn.base import clone

    op_compat = make_sklearn_compat(op)

    wrapped_op = clone(op_compat)
    cfg2 = smac_fixup_params(cfg)
    trainable = wrapped_op.set_params(**cfg2)
    return trainable

def get_smac_space(op:'Ops.PlannedOperator',
                   lale_num_grids:Optional[float]=None,
                   lale_pgo:Optional[PGO]=None)->ConfigurationSpace:
    """ Top level function: given a lale operator, returns a ConfigurationSpace for use with SMAC
    Parameters
    ----------
    op : The lale PlannedOperator
    lale_num_grids: integer or float, optional
        if set to an integer => 1, it will determine how many parameter grids will be returned (at most)
        if set to an float between 0 and 1, it will determine what fraction should be returned
        note that setting it to 1 is treated as in integer.  To return all results, use None
    """
    
    hp_grids = get_search_space_grids(op, num_grids=lale_num_grids, pgo=lale_pgo)
    cs = hp_grids_to_smac_cs(hp_grids)
    if should_print_search_space("true", "all", "backend", "smac"):
        name = op.name()
        if not name:
            name = "an operator"
        print(f"SMAC configuration for {name}:\n{str(cs)}")

    return cs

def smac_fixup_params(cfg):
    def strip_key(k:str)->str:
        return k.rsplit('_', 1)[0]
    def transform_value(v):
        if v == '_lale_none':
            return None
        else:
            return v

    ret = {strip_key(k):transform_value(v) for (k,v) in cfg.get_dictionary().items() if k != 'disjunct_discriminant'}
    return ret

# When sampling from distributions, this is the default number of samples to take.
# Users can override this by passing in num_samples to the appropriate function
SAMPLES_PER_DISTRIBUTION=2

# We can first convert from our search space IR
# to a more limited grid structure
# This can than be converted to the format required for SMAC

def SearchSpaceNumberToSMAC(key:str, hp:SearchSpaceNumber)->Hyperparameter:
    """Returns either a list of values intended to be sampled uniformly or a frozen scipy.stats distribution
    """
    dist = "uniform"
    if hp.distribution:
        dist = hp.distribution    
    if hp.maximum is None:
        raise ValueError(f"maximum not specified for a number with distribution {dist} for {key}")
    max = hp.getInclusiveMax()
    if hp.minimum is None:
        raise ValueError(f"minimum not specified for a number with distribution {dist} for {key}")
    min = hp.getInclusiveMin()

    log:bool
    if dist == "uniform" or dist == "integer":
        log = False
    elif dist == "loguniform":
        log = True
    else:
        raise ValueError(f"unknown/unsupported distribution {dist} for {key}")

    if hp.discrete:
        return UniformIntegerHyperparameter(key, min, max, log=log)
    else:
        return UniformFloatHyperparameter(key, min, max, log=log)

class FakeNone(object):
    pass

MyFakeNone = FakeNone()

def HPValuetoSMAC(key:str, hp:SearchSpace)->Hyperparameter:
    def val_to_str(v):
        if v is None:
            return "_lale_none"
        else:
            return v

    if isinstance(hp, SearchSpaceEnum):
        return CategoricalHyperparameter(key, list(map(val_to_str, hp.vals)))
    elif isinstance(hp, SearchSpaceNumber):
        return SearchSpaceNumberToSMAC(key, hp)
    elif isinstance(hp, SearchSpaceArray):
        raise ValueError(f"Arrays are not yet supported by the SMAC backend (key: {key})")
    else:
        raise ValueError(f"Not yet supported hp description ({type(hp)}) (key: {key}) in the GridSearchCV backend")
    
def SearchSpaceGridtoSMAC(hp:SearchSpaceGrid, disc:int)->Iterable[Hyperparameter]:
    return (HPValuetoSMAC(f"{k}_{disc}" , v) for k,v in hp.items())

disc_str = "disjunct_discriminant"
def addSearchSpaceGrid(hp:SearchSpaceGrid, disc:int, parent_disc:Hyperparameter, cs:ConfigurationSpace)->None:
    smac = SearchSpaceGridtoSMAC(hp, disc)
    for hyp in smac:
        cs.add_hyperparameter(hyp)
        cs.add_condition(EqualsCondition(child=hyp, parent=parent_disc, value=disc))

def addSearchSpaceGrids(grids:List[SearchSpaceGrid], cs:ConfigurationSpace)->None:
    parent_disc = CategoricalHyperparameter(disc_str, range(len(grids)))
    cs.add_hyperparameter(parent_disc)
    for (i,g) in enumerate(grids):
        addSearchSpaceGrid(g, i, parent_disc, cs)

def hp_grids_to_smac_cs(grids:List[SearchSpaceGrid])->ConfigurationSpace:
    cs:ConfigurationSpace = ConfigurationSpace()
    addSearchSpaceGrids(grids, cs)
    return cs




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from imblearn.under_sampling import EditedNearestNeighbours as OrigModel
import lale.operators 
from lale.lib.imblearn.base_resampler import BaseResamplerImpl, _input_fit_schema,\
                                            _input_transform_schema, _output_transform_schema,\
                                            _input_predict_schema, _output_predict_schema,\
                                            _input_predict_proba_schema, _output_predict_proba_schema,\
                                            _input_decision_function_schema, _output_decision_function_schema

class EditedNearestNeighboursImpl(BaseResamplerImpl):

    def __init__(self, operator = None, sampling_strategy='auto', random_state=None, 
                n_neighbors=3, kind_sel='all', n_jobs=1):
        if operator is None:
            raise ValueError("Operator is a required argument.")

        self._hyperparams = {
            'sampling_strategy': sampling_strategy,
            'random_state': random_state,
            'n_neighbors': n_neighbors,
            'kind_sel': kind_sel,
            'n_jobs': n_jobs}
    
        resampler_instance = OrigModel(**self._hyperparams)
        super(EditedNearestNeighboursImpl, self).__init__(
            operator = operator,
            resampler = resampler_instance)

_hyperparams_schema = {
    'allOf': [
    {   'type': 'object',
        'relevantToOptimizer': ['operator'],
        'additionalProperties': False,
        'properties': {
            'operator':{
                'description': """Trainable Lale pipeline that is trained using the data obtained from the current imbalance corrector.
Predict, transform, predict_proba or decision_function would just be forwarded to the trained pipeline.
If operator is a Planned pipeline, the current imbalance corrector can't be trained without using an optimizer to 
choose a trainable operator first. Please refer to lale/examples for more examples.""",
                'anyOf': [
                {   'laleType': 'operator'}]},
            'sampling_strategy': {
                'description': """sampling_strategy : str, list or callable, default='auto'. 
Sampling information to resample the data set.
""",
                'anyOf': [
                    {   'description':"""When ``str``, specify the class targeted by the resampling. 
The number of samples in the different classes will be equalized.
Possible choices are:
``'minority'``: resample only the minority class;
``'not minority'``: resample all classes but the minority class;
``'not majority'``: resample all classes but the majority class;
``'all'``: resample all classes;
``'auto'``: equivalent to ``'not majority'``.""",
                        'enum': ['minority','not minority','not majority', 'all', 'auto']},
                    {   'description':"""- When ``list``, the list contains the classes targeted by the resampling.""",
                        'anyOf': [
                            {'type': 'array', 'items': {'type': 'number'}},
                            {'type': 'array', 'items': {'type': 'string'}}]},
                    {   'description':"""When callable, function taking ``y`` and returns a ``dict``. 
The keys correspond to the targeted classes. The values correspond to the
desired number of samples for each class.""",
                        'laleType': 'Any'}],
                'default': 'auto'},
            'random_state': {
            'description':
                'Control the randomization of the algorithm.',
            'anyOf': [
                { 'description': 'RandomState used by np.random',
                'enum': [None]},
                { 'description': 'The seed used by the random number generator',
                'type': 'integer'},
                { 'description': 'Random number generator instance.',
                'laleType':'Any'}],
            'default': None},
            'n_neighbors':{
                'description': """If ``int``, size of the neighbourhood to consider to compute the nearest neighbors.  
If object, an estimator that inherits from
:class:`sklearn.neighbors.base.KNeighborsMixin` that will be used to
find the nearest-neighbors. Default of None corresponds to KNeighborsClassifier(n_neighbors=1)""",
                'anyOf': [
                    {'laleType':'Any'},
                    {'type': 'integer'}],
                'default': 3},
            'kind_sel':{
                'description':"""Strategy to use in order to exclude samples.
If ``all``, all neighbours will have to agree with the samples of interest to not be excluded.
If ``mode``, the majority vote of the neighbours will be used in order to exclude a sample.""",
                'enum':['all', 'mode'],
                'default':'all'},
            'n_jobs': {
                'description': 'The number of threads to open if possible.',
                'type': 'integer',
                'default': 1}}}]}

_combined_schemas = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': """Class to perform under-sampling based on the edited nearest neighbour method.""",
  'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.imblearn.edited_nearest_neighbours.html',
  'type': 'object',
  'tags': {
    'pre': [],
    'op': ['transformer', 'estimator', 'resampler'],#transformer and estimator both as a higher-order operator
    'post': []},
  'properties': {
    'hyperparams': _hyperparams_schema,
    'input_fit': _input_fit_schema,
    'input_transform': _input_transform_schema,
    'output_transform': _output_transform_schema,
    'input_predict': _input_predict_schema,
    'output_predict': _output_predict_schema,
    'input_predict_proba': _input_predict_proba_schema,
    'output_predict_proba': _output_predict_proba_schema,
    'input_decision_function': _input_decision_function_schema,
    'output_decision_function': _output_decision_function_schema
}}

lale.docstrings.set_docstrings(EditedNearestNeighboursImpl, _combined_schemas)

EditedNearestNeighbours = lale.operators.make_operator(EditedNearestNeighboursImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.type_checking
import numpy as np
import pandas as pd
import scipy.sparse
try:
    import torch
    torch_installed=True
except ImportError:
    torch_installed=False
# See instructions for subclassing numpy ndarray:
# https://docs.scipy.org/doc/numpy/user/basics.subclassing.html
class NDArrayWithSchema(np.ndarray):
    def __new__(subtype, shape, dtype=float, buffer=None, offset=0,
                strides=None, order=None, json_schema=None):
        result = super(NDArrayWithSchema, subtype).__new__(
            subtype, shape, dtype, buffer, offset, strides, order)
        result.json_schema = json_schema
        return result

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.json_schema = getattr(obj, 'json_schema', None)

# See instructions for subclassing pandas DataFrame:
# https://pandas.pydata.org/pandas-docs/stable/development/extending.html#extending-subclassing-pandas
class DataFrameWithSchema(pd.DataFrame):
    _internal_names = pd.DataFrame._internal_names + ['json_schema']
    _internal_names_set = set(_internal_names)

    @property
    def _constructor(self):
        return DataFrameWithSchema

class SeriesWithSchema(pd.Series):
    _internal_names = pd.Series._internal_names + ['json_schema']
    _internal_names_set = set(_internal_names)

    @property
    def _constructor(self):
        return SeriesWithSchema

def is_list_tensor(obj):
    def list_tensor_shape(ls):
        if isinstance(ls, int) or isinstance(ls, float) or isinstance(ls, str):
            return (str(type(ls)),)
        if isinstance(ls, list):
            sub_shape = 'Any'
            for item in ls:
                item_result = list_tensor_shape(item)
                if item_result is None:
                    return None
                if sub_shape == 'Any':
                    sub_shape = item_result
                elif sub_shape != item_result:
                    return None
            return (len(ls),) + sub_shape
        return None
    if isinstance(obj, list):
        shape = list_tensor_shape(obj)
        return shape is not None
    return False

def add_schema(obj, schema=None, raise_on_failure=False, recalc=False):
    if obj is None:
        return None
    if isinstance(obj, NDArrayWithSchema):
        result = obj
    elif isinstance(obj, np.ndarray):
        result = obj.view(NDArrayWithSchema)
    elif isinstance(obj, SeriesWithSchema):
        result = obj
    elif isinstance(obj, pd.Series):
        result = SeriesWithSchema(obj)
    elif isinstance(obj, DataFrameWithSchema):
        result = obj
    elif isinstance(obj, pd.DataFrame):
        result = DataFrameWithSchema(obj)
    elif is_list_tensor(obj):
        obj = np.array(obj)
        result = obj.view(NDArrayWithSchema)
    elif raise_on_failure:
        raise ValueError(f'unexpected type(obj) {type(obj)}')
    else:
        return obj
    if recalc:
        result.json_schema = None
    if not hasattr(result, 'json_schema') or result.json_schema is None:
        if schema is None:
            result.json_schema = to_schema(obj)
        else:
            lale.type_checking.validate_is_schema(schema)
            result.json_schema = schema
    return result

def dtype_to_schema(typ):
    result = None
    if typ is bool or typ is np.bool_:
        result = {'type': 'boolean'}
    elif np.issubdtype(typ, np.unsignedinteger):
        result = {'type': 'integer', 'minimum': 0}
    elif np.issubdtype(typ, np.integer):
        result = {'type': 'integer'}
    elif np.issubdtype(typ, np.number):
        result = {'type': 'number'}
    elif np.issubdtype(typ, np.string_) or np.issubdtype(typ, np.unicode_):
        result = {'type': 'string'}
    elif isinstance(typ, np.dtype):
        if typ.fields:
            props = {k: dtype_to_schema(t) for k, t in typ.fields.items()}
            result = {'type': 'object', 'properties': props}
        elif typ.shape:
            result = shape_and_dtype_to_schema(typ.shape, typ.subdtype)
        elif np.issubdtype(typ, np.object_):
            result = {'type': 'string'}
        else:
            assert False, f'unexpected dtype {typ}'
    else:
        assert False, f'unexpected non-dtype {typ}'
    lale.type_checking.validate_is_schema(result)
    return result

def shape_and_dtype_to_schema(shape, dtype):
    result = dtype_to_schema(dtype)
    for dim in reversed(shape):
        result = {
            'type': 'array',
            'minItems': dim,
            'maxItems': dim,
            'items': result}
    lale.type_checking.validate_is_schema(result)
    return result

def ndarray_to_schema(array):
    assert isinstance(array, np.ndarray)
    if isinstance(array, NDArrayWithSchema) and hasattr(array, 'json_schema') and array.json_schema is not None:
        return array.json_schema
    return shape_and_dtype_to_schema(array.shape, array.dtype)

def csr_matrix_to_schema(matrix):
    assert isinstance(matrix, scipy.sparse.csr_matrix)
    return shape_and_dtype_to_schema(matrix.shape, matrix.dtype)

def dataframe_to_schema(df):
    assert isinstance(df, pd.DataFrame)
    if isinstance(df, DataFrameWithSchema) and hasattr(df, 'json_schema') and df.json_schema is not None:
        return df.json_schema
    n_rows, n_columns = df.shape
    assert n_columns == len(df.columns) and n_columns == len(df.dtypes)
    items = [
        {'description': str(col), **dtype_to_schema(df.dtypes[col])}
        for col in df.columns]
    result = {
        'type': 'array',
        'minItems': n_rows,
        'maxItems': n_rows,
        'items': {
            'type': 'array',
            'minItems': n_columns,
            'maxItems': n_columns,
            'items': items}}
    lale.type_checking.validate_is_schema(result)
    return result

def series_to_schema(series):
    assert isinstance(series, pd.Series)
    if isinstance(series, SeriesWithSchema) and hasattr(series, 'json_schema') and series.json_schema is not None:
        return series.json_schema
    (n_rows, ) = series.shape
    result = {
        'type': 'array',
        'minItems': n_rows,
        'maxItems': n_rows,
        'items': {
            'description': str(series.name),
            **dtype_to_schema(series.dtype)}}
    lale.type_checking.validate_is_schema(result)
    return result

def torch_tensor_to_schema(tensor):
    assert torch_installed, """Your Python environment does not have torch installed. You can install it with
    pip install torch
or with
    pip install 'lale[full]'"""
    assert isinstance(tensor, torch.Tensor)
    #https://pytorch.org/docs/stable/tensor_attributes.html#torch-dtype
    if tensor.dtype == torch.bool:
        result = {'type': 'boolean'}
    elif tensor.dtype == torch.uint8:
        result = {'type': 'integer', 'minimum': 0, 'maximum': 255}
    elif torch.is_floating_point(tensor):
        result = {'type': 'number'}
    else:
        result = {'type': 'integer'}
    for dim in reversed(tensor.shape):
        result = {
            'type': 'array',
            'minItems': dim, 'maxItems': dim,
            'items': result}
    return result

def is_liac_arff(obj):
    expected_types = {
        'description': str, 'relation': str, 'attributes': list, 'data': list}
    if not isinstance(obj, dict):
        return False
    for k, t in expected_types.items():
        if k not in obj or not isinstance(obj[k], t):
            return False
    return True

def liac_arff_to_schema(larff):
    assert is_liac_arff(larff), """Your Python environment might contain an 'arff' package different from 'liac-arff'. You can install it with
    pip install 'liac-arff>=2.4.0'
or with
    pip install 'lale[full]'"""
    n_rows, n_columns = len(larff['data']), len(larff['attributes'])
    def larff_type_to_schema(larff_type):
        if isinstance(larff_type, str):
            a2j = {'numeric': 'number', 'real': 'number',
                   'integer': 'integer', 'string': 'string'}
            return {'type': a2j[larff_type.lower()]}
        assert isinstance(larff_type, list)
        return {'enum': [*larff_type]}
    items = [
        {'description': attr[0], **larff_type_to_schema(attr[1])}
        for attr in larff['attributes']]
    result = {
        'type': 'array',
        'minItems': n_rows,
        'maxItems': n_rows,
        'items': {
            'type': 'array',
            'minItems': n_columns,
            'maxItems': n_columns,
            'items': items}}
    lale.type_checking.validate_is_schema(result)
    return result

def to_schema(obj):
    if obj is None:
        result = {'enum': [None]}
    elif isinstance(obj, np.ndarray):
        result = ndarray_to_schema(obj)
    elif isinstance(obj, scipy.sparse.csr_matrix):
        result = csr_matrix_to_schema(obj)
    elif isinstance(obj, pd.DataFrame):
        result = dataframe_to_schema(obj)
    elif isinstance(obj, pd.Series):
        result = series_to_schema(obj)
    elif torch_installed and isinstance(obj, torch.Tensor):
        result = torch_tensor_to_schema(obj)
    elif is_liac_arff(obj):
        result = liac_arff_to_schema(obj)
    elif lale.type_checking.is_schema(obj):
        result = obj
    else:
        raise ValueError(f'to_schema(obj), type {type(obj)}, value {obj}')
    lale.type_checking.validate_is_schema(result)
    return result



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.discriminant_analysis
import lale.docstrings
import lale.operators

class QuadraticDiscriminantAnalysisImpl():

    def __init__(self, priors=None, reg_param=None, store_covariance=None, tol=0.0001, store_covariances=None):
        self._hyperparams = {
            'priors': priors,
            'reg_param': reg_param,
            'store_covariance': store_covariance,
            'tol': tol,
            'store_covariances': store_covariances}
        self._wrapped_model = sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis(**self._hyperparams)

    def fit(self, X, y=None):
        self._wrapped_model.fit(X, y)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)

    def decision_function(self, X):
        return self._wrapped_model.decision_function(X)

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Quadratic Discriminant Analysis',
    'allOf': [{
        'type': 'object',
        'required': ['priors', 'store_covariance'],
        'relevantToOptimizer': ['reg_param', 'tol'],
        'additionalProperties': False,
        'properties': {
            'priors': {
                'anyOf': [{
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }, {
                    'enum': [None]}],
                'default': None,
                'description': 'Priors on classes'},
            'reg_param': {
                'type': 'number',
                'minimumForOptimizer': 0.0,
                'maximumForOptimizer': 1.0,
                'distribution': 'uniform',
                'default': 0.0,
                'description': 'Regularizes the covariance estimate as'},
            'store_covariance': {
                'type': 'boolean',
                'default': False,
                'description': 'If True the covariance matrices are computed and stored in the'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'loguniform',
                'default': 0.0001,
                'description': 'Threshold used for rank estimation.'}
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the model according to the given training data and parameters.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vector, where n_samples is the number of samples and'},
        'y': {
            'anyOf': [
                {'type': 'array', 'items': {'type': 'number'}},
                {'type': 'array', 'items': {'type': 'string'}}],
            'description': 'Target values (integers)'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Perform classification on an array of test vectors X.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Perform classification on an array of test vectors X.',
    'anyOf': [
        {'type': 'array', 'items': {'type': 'number'}},
        {'type': 'array', 'items': {'type': 'string'}}]}

_input_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Return posterior probabilities of classification.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Array of samples/test vectors.'},
    },
}
_output_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Posterior probabilities of classification per class.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}

_input_decision_function_schema = {
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}}}}

_output_decision_function_schema = {
    'description': 'Confidence scores for samples for each class in the model.',
    'anyOf': [
    {   'description': 'In the multi-way case, score per (sample, class) combination.',
        'type': 'array',
        'items': {'type': 'array', 'items': {'type': 'number'}}},
    {   'description': 'In the binary case, score for `self._classes[1]`.',
        'type': 'array',
        'items': {'type': 'number'}}]}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis.html#sklearn-discriminant-analysis-quadraticdiscriminantanalysis',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'classifier'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_predict_proba': _input_predict_proba_schema,
        'output_predict_proba': _output_predict_proba_schema,
        'input_decision_function': _input_decision_function_schema,
        'output_decision_function': _output_decision_function_schema,
}}

lale.docstrings.set_docstrings(QuadraticDiscriminantAnalysisImpl, _combined_schemas)

QuadraticDiscriminantAnalysis = lale.operators.make_operator(QuadraticDiscriminantAnalysisImpl, _combined_schemas)




from sklearn.cluster.hierarchical import FeatureAgglomeration as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class FeatureAgglomerationImpl():

    def __init__(self, n_clusters=2, affinity='euclidean', memory=None, connectivity=None, compute_full_tree='auto', linkage='ward', pooling_func=None):
        self._hyperparams = {
            'n_clusters': n_clusters,
            'affinity': affinity,
            'memory': memory,
            'connectivity': connectivity,
            'compute_full_tree': compute_full_tree,
            'linkage': linkage,
            'pooling_func': pooling_func}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for FeatureAgglomeration    Agglomerate features.',
    'allOf': [{
        'type': 'object',
        'required': ['n_clusters', 'affinity', 'memory', 'connectivity', 'compute_full_tree', 'linkage', 'pooling_func'],
        'relevantToOptimizer': ['n_clusters', 'affinity', 'compute_full_tree', 'linkage'],
        'additionalProperties': False,
        'properties': {
            'n_clusters': {
                'type': 'integer',
                'minimumForOptimizer': 2,
                'maximumForOptimizer': 8,
                'distribution': 'uniform',
                'default': 2,
                'description': 'The number of clusters to find.'},
            'affinity': {
                'anyOf': [{
                    'type': 'object',
                    'forOptimizer': False}, {
                    'enum': ['euclidean']}],
                'default': 'euclidean',
                'description': 'Metric used to compute the linkage'},
            'memory': {
                'XXX TODO XXX': 'None, str or object with the joblib.Memory interface, optional',
                'description': 'Used to cache the output of the computation of the tree',
                'enum': [None],
                'default': None},
            'connectivity': {
                'anyOf': [{
                    'type': 'array',
                    'items': {
                        'laleType': 'Any',
                        'XXX TODO XXX': 'item type'},
                    'XXX TODO XXX': 'array-like or callable, optional'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Connectivity matrix'},
            'compute_full_tree': {
                'XXX TODO XXX': 'bool or \'auto\', optional, default "auto"',
                'description': 'Stop early the construction of the tree at n_clusters',
                'enum': ['auto'],
                'default': 'auto'},
            'linkage': {
                'enum': ['ward', 'complete', 'average', 'single'],
                'default': 'ward',
                'description': 'Which linkage criterion to use'},
            'pooling_func': {
                'laleType': 'Any',
                'XXX TODO XXX': 'callable, default np.mean',
                'description': 'This combines the values of agglomerated features into a single value, and should accept an array of shape [M, N] and the keyword argument `axis=1`, and reduce it to an array of size [M].'},
        }}, {
        'XXX TODO XXX': 'Parameter: affinity > only "euclidean" is accepted'}, {
        'XXX TODO XXX': 'Parameter: compute_full_tree > only when specifying a connectivity matrix'}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the hierarchical clustering on the data',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The data'},
        'y': {
            
        }},
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform a new matrix using the built clustering',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}, {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }],
            'description': 'A M by N array of M observations in N dimensions or a length M array of M one-dimensional observations.'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The pooled values for each feature cluster.',
    'anyOf': [{
        'type': 'array',
        'items': {
            'type': 'array',
            'items': {
                'type': 'number'},
        }}, {
        'type': 'array',
        'items': {
            'type': 'number'},
    }],
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.cluster.FeatureAgglomeration#sklearn-cluster-featureagglomeration',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(FeatureAgglomerationImpl, _combined_schemas)
FeatureAgglomeration = lale.operators.make_operator(FeatureAgglomerationImpl, _combined_schemas)





from sklearn.linear_model.coordinate_descent import MultiTaskElasticNet as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class MultiTaskElasticNetImpl():

    def __init__(self, alpha=1.0, l1_ratio=0.5, fit_intercept=True, normalize=False, copy_X=True, max_iter=1000, tol=0.0001, warm_start=False, random_state=None, selection='cyclic'):
        self._hyperparams = {
            'alpha': alpha,
            'l1_ratio': l1_ratio,
            'fit_intercept': fit_intercept,
            'normalize': normalize,
            'copy_X': copy_X,
            'max_iter': max_iter,
            'tol': tol,
            'warm_start': warm_start,
            'random_state': random_state,
            'selection': selection}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for MultiTaskElasticNet    Multi-task ElasticNet model trained with L1/L2 mixed-norm as regularizer',
    'allOf': [{
        'type': 'object',
        'required': ['alpha', 'l1_ratio', 'fit_intercept', 'normalize', 'copy_X', 'max_iter', 'tol', 'warm_start', 'random_state', 'selection'],
        'relevantToOptimizer': ['alpha', 'fit_intercept', 'normalize', 'copy_X', 'max_iter', 'tol'],
        'additionalProperties': False,
        'properties': {
            'alpha': {
                'type': 'number',
                'minimumForOptimizer': 1e-10,
                'maximumForOptimizer': 1.0,
                'distribution': 'uniform',
                'default': 1.0,
                'description': 'Constant that multiplies the L1/L2 term'},
            'l1_ratio': {
                'type': 'number',
                'default': 0.5,
                'description': 'The ElasticNet mixing parameter, with 0 < l1_ratio <= 1'},
            'fit_intercept': {
                'type': 'boolean',
                'default': True,
                'description': 'whether to calculate the intercept for this model'},
            'normalize': {
                'type': 'boolean',
                'default': False,
                'description': 'This parameter is ignored when ``fit_intercept`` is set to False'},
            'copy_X': {
                'type': 'boolean',
                'default': True,
                'description': 'If ``True``, X will be copied; else, it may be overwritten.'},
            'max_iter': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 1000,
                'description': 'The maximum number of iterations'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'uniform',
                'default': 0.0001,
                'description': 'The tolerance for the optimization: if the updates are smaller than ``tol``, the optimization code checks the dual gap for optimality and continues until it is smaller than ``tol``.'},
            'warm_start': {
                'type': 'boolean',
                'default': False,
                'description': 'When set to ``True``, reuse the solution of the previous call to fit as initialization, otherwise, just erase the previous solution'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The seed of the pseudo random number generator that selects a random feature to update'},
            'selection': {
                'type': 'string',
                'default': 'cyclic',
                'description': "If set to 'random', a random coefficient is updated every iteration rather than looping over features sequentially by default"},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit MultiTaskElasticNet model with coordinate descent',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'laleType': 'Any',
            'XXX TODO XXX': 'ndarray, shape (n_samples, n_features)',
            'description': 'Data'},
        'y': {
            'laleType': 'Any',
            'XXX TODO XXX': 'ndarray, shape (n_samples, n_tasks)',
            'description': 'Target'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict using the linear model',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array_like or sparse matrix, shape (n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Samples.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Returns predicted values.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.MultiTaskElasticNet#sklearn-linear_model-multitaskelasticnet',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema},
}
lale.docstrings.set_docstrings(MultiTaskElasticNetImpl, _combined_schemas)
MultiTaskElasticNet = lale.operators.make_operator(MultiTaskElasticNetImpl, _combined_schemas)





from sklearn.linear_model.ridge import RidgeClassifierCV as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class RidgeClassifierCVImpl():

    def __init__(self, alphas=[0.1, 1.0, 10.0], fit_intercept=True, normalize=False, scoring=None, cv=None, class_weight='balanced', store_cv_values=False):
        self._hyperparams = {
            'alphas': alphas,
            'fit_intercept': fit_intercept,
            'normalize': normalize,
            'scoring': scoring,
            'cv': cv,
            'class_weight': class_weight,
            'store_cv_values': store_cv_values}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def decision_function(self, X):
        return self._wrapped_model.decision_function(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for RidgeClassifierCV    Ridge classifier with built-in cross-validation.',
    'allOf': [{
        'type': 'object',
        'required': ['alphas', 'fit_intercept', 'normalize', 'scoring', 'cv', 'class_weight', 'store_cv_values'],
        'relevantToOptimizer': ['fit_intercept', 'normalize', 'cv'],
        'additionalProperties': False,
        'properties': {
            'alphas': {
                'type': 'array',
                'items': {
                    'type': 'number'},
                'default': [0.1, 1.0, 10.0],
                'description': 'Array of alpha values to try'},
            'fit_intercept': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether to calculate the intercept for this model'},
            'normalize': {
                'type': 'boolean',
                'default': False,
                'description': 'This parameter is ignored when ``fit_intercept`` is set to False'},
            'scoring': {
                'anyOf': [{
                    'type': 'string'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'A string (see model evaluation documentation) or a scorer callable object / function with signature ``scorer(estimator, X, y)``.'},
            'cv': {
                'XXX TODO XXX': 'int, cross-validation generator or an iterable, optional',
                'description': 'Determines the cross-validation splitting strategy',
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 3,
                    'maximumForOptimizer': 4,
                    'distribution': 'uniform'}, {
                    'enum': [None]}],
                'default': None},
            'class_weight': {
                'XXX TODO XXX': "dict or 'balanced', optional",
                'description': 'Weights associated with classes in the form ``{class_label: weight}``',
                'enum': ['balanced'],
                'default': 'balanced'},
            'store_cv_values': {
                'type': 'boolean',
                'default': False,
                'description': 'Flag indicating if the cross-validation values corresponding to each alpha should be stored in the ``cv_values_`` attribute (see below)'},
        }}, {
        'XXX TODO XXX': 'Parameter: store_cv_values > only compatible with cv=none (i'}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the ridge classifier.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vectors, where n_samples is the number of samples and n_features is the number of features.'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target values'},
        'sample_weight': {
            'anyOf': [{
                'type': 'number'}, {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }],
            'description': 'Sample weight.'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict class labels for samples in X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array_like or sparse matrix, shape (n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Samples.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predicted class label per sample.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_input_decision_function_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict confidence scores for samples.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array_like or sparse matrix, shape (n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Samples.'},
    },
}
_output_decision_function_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Confidence scores per (sample, class) combination',
    'laleType': 'Any',
    'XXX TODO XXX': 'array, shape=(n_samples,) if n_classes == 2 else (n_samples, n_classes)',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.RidgeClassifierCV#sklearn-linear_model-ridgeclassifiercv',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_decision_function': _input_decision_function_schema,
        'output_decision_function': _output_decision_function_schema},
}
lale.docstrings.set_docstrings(RidgeClassifierCVImpl, _combined_schemas)
RidgeClassifierCV = lale.operators.make_operator(RidgeClassifierCVImpl, _combined_schemas)




# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import autoai_libs.transformers.exportable
import lale.datasets.data_schemas
import lale.docstrings
import lale.operators

class boolean2floatImpl():
    def __init__(self, activate_flag):
        self._hyperparams = {
            'activate_flag': activate_flag}
        self._autoai_tfm = autoai_libs.transformers.exportable.boolean2float(**self._hyperparams)

    def fit(self, X, y=None):
        self._autoai_tfm.fit(X, y)
        return self

    def transform(self, X):
        raw = self._autoai_tfm.transform(X)
        s_X = lale.datasets.data_schemas.to_schema(X)
        s_result = self.transform_schema(s_X)
        result = lale.datasets.data_schemas.add_schema(raw, s_result, recalc=True)
        assert result.json_schema == s_result
        return result

    def transform_schema(self, s_X):
        """Used internally by Lale for type-checking downstream operators."""
        if self._hyperparams['activate_flag']:
            result = {
                'type': 'array',
                'items': {'type': 'array', 'items': {'type': 'number'}}}
        else:
            result = s_X
        return result

_hyperparams_schema = {
    'allOf': [{
        'description': 'This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.',
        'type': 'object',
        'additionalProperties': False,
        'required': ['activate_flag'],
        'relevantToOptimizer': [],
        'properties': {
            'activate_flag': {
                'description': 'If False, transform(X) outputs the input numpy array X unmodified.',
                'type': 'boolean',
                'default': True}}}]}

_input_fit_schema = {
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {#Handles 1-D arrays as well
            'anyOf': [
                {'type': 'array', 'items': {'laleType': 'Any'}},
                {'type': 'array', 'items': {'type': 'array', 'items': {'laleType': 'Any'}}}]},
        'y': {
            'laleType': 'Any'}}}

_input_transform_schema = {
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {#Handles 1-D arrays as well
            'anyOf': [
                {'type': 'array', 'items': {'laleType': 'Any'}},
                {'type': 'array', 'items': {'type': 'array', 'items': {'laleType': 'Any'}}}]}}}

_output_transform_schema = {
    'description': 'Features; the outer array is over samples.',
    'type': 'array',
    'items': {'type': 'array', 'items': {'laleType': 'Any'}}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """Operator from `autoai_libs`_. Converts strings that represent booleans to floats and replaces missing values with np.nan.

.. _`autoai_libs`: https://pypi.org/project/autoai-libs""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.autoai_libs.boolean2float.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(boolean2floatImpl, _combined_schemas)

boolean2float = lale.operators.make_operator(boolean2floatImpl, _combined_schemas)




from sklearn.decomposition.online_lda import LatentDirichletAllocation as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class LatentDirichletAllocationImpl():

    def __init__(self, n_components=10, doc_topic_prior=None, topic_word_prior=None, learning_method='batch', learning_decay=0.7, learning_offset=10.0, max_iter=10, batch_size=128, evaluate_every=(- 1), total_samples=1000000.0, perp_tol=0.1, mean_change_tol=0.001, max_doc_update_iter=100, n_jobs=None, verbose=0, random_state=None, n_topics=None):
        self._hyperparams = {
            'n_components': n_components,
            'doc_topic_prior': doc_topic_prior,
            'topic_word_prior': topic_word_prior,
            'learning_method': learning_method,
            'learning_decay': learning_decay,
            'learning_offset': learning_offset,
            'max_iter': max_iter,
            'batch_size': batch_size,
            'evaluate_every': evaluate_every,
            'total_samples': total_samples,
            'perp_tol': perp_tol,
            'mean_change_tol': mean_change_tol,
            'max_doc_update_iter': max_doc_update_iter,
            'n_jobs': n_jobs,
            'verbose': verbose,
            'random_state': random_state,
            'n_topics': n_topics}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for LatentDirichletAllocation    Latent Dirichlet Allocation with online variational Bayes algorithm',
    'allOf': [{
        'type': 'object',
        'required': ['n_components', 'doc_topic_prior', 'topic_word_prior', 'learning_method', 'learning_decay', 'learning_offset', 'max_iter', 'batch_size', 'evaluate_every', 'total_samples', 'perp_tol', 'mean_change_tol', 'max_doc_update_iter', 'n_jobs', 'verbose', 'random_state', 'n_topics'],
        'relevantToOptimizer': ['n_components', 'learning_method', 'max_iter', 'batch_size', 'evaluate_every', 'total_samples', 'max_doc_update_iter'],
        'additionalProperties': False,
        'properties': {
            'n_components': {
                'type': 'integer',
                'minimumForOptimizer': 2,
                'maximumForOptimizer': 256,
                'distribution': 'uniform',
                'default': 10,
                'description': 'Number of topics.'},
            'doc_topic_prior': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Prior of document topic distribution `theta`'},
            'topic_word_prior': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Prior of topic word distribution `beta`'},
            'learning_method': {
                'enum': ['batch', 'online'],
                'default': 'batch',
                'description': 'Method used to update `_component`'},
            'learning_decay': {
                'type': 'number',
                'default': 0.7,
                'description': 'It is a parameter that control learning rate in the online learning method'},
            'learning_offset': {
                'type': 'number',
                'default': 10.0,
                'description': 'A (positive) parameter that downweights early iterations in online learning'},
            'max_iter': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 10,
                'description': 'The maximum number of iterations.'},
            'batch_size': {
                'type': 'integer',
                'minimumForOptimizer': 3,
                'maximumForOptimizer': 128,
                'distribution': 'uniform',
                'default': 128,
                'description': 'Number of documents to use in each EM iteration'},
            'evaluate_every': {
                'type': 'integer',
                'minimumForOptimizer': (- 1),
                'maximumForOptimizer': 0,
                'distribution': 'uniform',
                'default': (- 1),
                'description': 'How often to evaluate perplexity'},
            'total_samples': {
                'anyOf': [{
                    'type': 'integer',
                    'forOptimizer': False}, {
                    'type': 'number',
                    'minimumForOptimizer': 0.0,
                    'maximumForOptimizer': 1.0,
                    'distribution': 'loguniform'}],
                'default': 1000000.0,
                'description': 'Total number of documents'},
            'perp_tol': {
                'type': 'number',
                'default': 0.1,
                'description': 'Perplexity tolerance in batch learning'},
            'mean_change_tol': {
                'type': 'number',
                'default': 0.001,
                'description': 'Stopping tolerance for updating document topic distribution in E-step.'},
            'max_doc_update_iter': {
                'type': 'integer',
                'minimumForOptimizer': 100,
                'maximumForOptimizer': 101,
                'distribution': 'uniform',
                'default': 100,
                'description': 'Max number of iterations for updating document topic distribution in the E-step.'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of jobs to use in the E-step'},
            'verbose': {
                'type': 'integer',
                'default': 0,
                'description': 'Verbosity level.'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`.'},
            'n_topics': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'This parameter has been renamed to n_components and will be removed in version 0.21'},
        }}, {
        'description': 'learning_method, only used in fit method',
        'anyOf': [{
            'type': 'object',
            'properties': {
                'learning_method': {
                    'enum': ['batch']},
            }}, {
            'type': 'object',
            'properties': {
                'method': {
                    'enum': ['fit']},
            }}]}, {
        'description': 'batch_size, only used in online learning',
        'anyOf': [{
            'type': 'object',
            'properties': {
                'batch_size': {
                    'enum': [128]},
            }}, {
            'type': 'object',
            'properties': {
                'learning': {
                    'enum': ['online']},
            }}]}, {
        'description': 'evaluate_every, only used in fit method',
        'anyOf': [{
            'type': 'object',
            'properties': {
                'evaluate_every': {
                    'enum': [(- 1)]},
            }}, {
            'type': 'object',
            'properties': {
                'method': {
                    'enum': ['fit']},
            }}]}, {
        'description': 'total_samples, only used in the partial_fit method',
        'anyOf': [{
            'type': 'object',
            'properties': {
                'total_samples': {
                    'enum': [1000000.0]},
            }}, {
            'type': 'object',
            'properties': {
                'method': {
                    'enum': ['partial_fit']},
            }}]}, {
        'XXX TODO XXX': 'Parameter: perp_tol > only used when evaluate_every is greater than 0'}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Learn model for the data X with variational Bayes method.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array-like or sparse matrix, shape=(n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Document word matrix.'},
        'y': {
            
        }},
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform data X according to the fitted model.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array-like or sparse matrix, shape=(n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Document word matrix.'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Document topic distribution for X.',
    'laleType': 'Any',
    'XXX TODO XXX': 'shape=(n_samples, n_components)',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.decomposition.LatentDirichletAllocation#sklearn-decomposition-latentdirichletallocation',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(LatentDirichletAllocationImpl, _combined_schemas)
LatentDirichletAllocation = lale.operators.make_operator(LatentDirichletAllocationImpl, _combined_schemas)

